@@ -6,7 +6,7 @@ Redis::JobQueue - Job queue management implemented using Redis server.
 
 =head1 VERSION
 
-This documentation refers to C<Redis::JobQueue> version 1.18
+This documentation refers to C<Redis::JobQueue> version 1.19
 
 =cut
 
@@ -18,7 +18,7 @@ use warnings;
 
 # ENVIRONMENT ------------------------------------------------------------------
 
-our $VERSION = '1.18';
+our $VERSION = '1.19';
 
 use Exporter qw(
     import
@@ -27,6 +27,8 @@ our @EXPORT_OK  = qw(
     DEFAULT_SERVER
     DEFAULT_PORT
     DEFAULT_TIMEOUT
+    DEFAULT_CONNECTION_TIMEOUT
+    DEFAULT_OPERATION_TIMEOUT
     NAMESPACE
     E_NO_ERROR
     E_MISMATCH_ARG
@@ -73,6 +75,7 @@ use Storable qw(
     nfreeze
     thaw
 );
+use Time::HiRes ();
 use Try::Tiny;
 
 #-- declarations ---------------------------------------------------------------
@@ -255,6 +258,20 @@ Maximum time (in seconds) to wait for a new job from the queue,
 =cut
 use constant DEFAULT_TIMEOUT    => 0;           # 0 for an unlimited timeout
 
+=item C<DEFAULT_CONNECTION_TIMEOUT>
+
+Default socket timeout for connection, number of seconds: 0.1 .
+
+=cut
+use constant DEFAULT_CONNECTION_TIMEOUT => 0.1;
+
+=item C<DEFAULT_OPERATION_TIMEOUT>
+
+Default socket timeout for read and write operations, number of seconds: 1.
+
+=cut
+use constant DEFAULT_OPERATION_TIMEOUT  => 1;
+
 =item C<NAMESPACE>
 
 Namespace used for keys on the Redis server - C<'JobQueue'>.
@@ -331,11 +348,13 @@ our @ERROR = (
     'Data is too large',
     'Error in connection to Redis server',
     "Command not allowed when used memory > 'maxmemory'",
-    'job was removed by maxmemory-policy',
     'job was removed prior to use',
     'Redis error',
 );
 
+our $WAIT_USED_MEMORY   = 0;    # No attempt to protect against used_memory > maxmemory
+#our $WAIT_USED_MEMORY   = 1;
+
 my $_ID_IN_QUEUE_FIELD = '__id_in_queue__';
 
 my $NAMESPACE = NAMESPACE;
@@ -346,6 +365,7 @@ my %job_fnames = map { $_ => 1 } @job_fields;
 
 my $uuid = new Data::UUID;
 
+my ( $_running_script_name, $_running_script_body );
 my %lua_script_body;
 my $lua_id_rgxp = '([^:]+)$';
 
@@ -565,7 +585,7 @@ END_GET_JOB_IDS
 
 =head2 CONSTRUCTOR
 
-=head3 C<new( redis =E<gt> $server, timeout =E<gt> $timeout, check_maxmemory =E<gt> $mode )>
+=head3 C<new( redis =E<gt> $server, timeout =E<gt> $timeout, check_maxmemory =E<gt> $mode, ... )>
 
 Creates a new C<Redis::JobQueue> object to communicate with Redis server.
 If invoked without any arguments, the constructor C<new> creates and returns
@@ -581,6 +601,27 @@ defines if attempt is made to find out maximum available memory from Redis.
 In some cases Redis implementation forbids such request,
 but setting <check_maxmemory> to false can be used as a workaround.
 
+This example illustrates a C<new()> call with all the valid arguments:
+
+    my $coll = Redis::CappedCollection->create(
+        redis   => { server => "$server:$port" },   # Redis object
+                                    # or hash reference to parameters to create a new Redis object.
+        timeout => $timeout,        # wait time (in seconds)
+                                    # for blocking call of get_next_job.
+                                    # Set 0 for unlimited wait time
+        check_maxmemory => $mode,   # boolean argument (default is true)
+                                    # defines if attempt is made to find out
+                                    # maximum available memory from Redis.
+        reconnect_on_error  => 1,   # Boolean argument - default is true and conservative_reconnect is true.
+                                    # Controls ability to force re-connection with Redis on error.
+        connection_timeout  => DEFAULT_CONNECTION_TIMEOUT,  # Socket timeout for connection,
+                                    # number of seconds (can be fractional).
+                                    # NOTE: Changes external socket configuration.
+        operation_timeout   => DEFAULT_OPERATION_TIMEOUT,   # Socket timeout for read and write operations,
+                                    # number of seconds (can be fractional).
+                                    # NOTE: Changes external socket configuration.
+    );
+
 =head3 Caveats related to connection with Redis server
 
 =over 3
@@ -691,7 +732,13 @@ around BUILDARGS => sub {
         elsif ( ref( $redis ) eq 'HASH' ) {
             $args{_use_external_connection} = 0;
             my $conf = $redis;
-            $conf->{server} = DEFAULT_SERVER.':'.DEFAULT_PORT unless exists $conf->{server};
+            $conf->{server}                 = DEFAULT_SERVER.':'.DEFAULT_PORT   unless exists $conf->{server};
+            $conf->{reconnect}              = 1                                 unless exists $conf->{reconnect};
+            $conf->{every}                  = 1000                              unless exists $conf->{every};   # 1 ms
+            $conf->{conservative_reconnect} = 1                                 unless exists $conf->{conservative_reconnect};
+            $conf->{cnx_timeout}    = $args{connection_timeout} = $conf->{cnx_timeout}  // $args{connection_timeout}    // DEFAULT_CONNECTION_TIMEOUT;
+            $conf->{read_timeout}   = $args{operation_timeout}  = $conf->{read_timeout} // $args{operation_timeout}     // DEFAULT_OPERATION_TIMEOUT;
+            $conf->{write_timeout}  = $conf->{read_timeout};
             delete $args{redis};
             return $class->$orig(
                 redis   => $conf->{server},
@@ -702,6 +749,8 @@ around BUILDARGS => sub {
             return $class->$orig( %args );
         }
     }
+
+    return;
 };
 
 sub BUILD {
@@ -709,6 +758,12 @@ sub BUILD {
 
     $self->_redis( $self->_redis_constructor )
         unless ( $self->_redis );
+    $self->_redis->connect if
+           exists( $self->_redis->{no_auto_connect_on_new} )
+        && $self->_redis->{no_auto_connect_on_new}
+        && !$self->_redis->{sock}
+    ;
+
     if ( $self->_check_maxmemory ) {
         my ( undef, $max_datasize ) = $self->_call_redis( 'CONFIG', 'GET', 'maxmemory' );
         defined( _NONNEGINT( $max_datasize ) )
@@ -717,15 +772,26 @@ sub BUILD {
             if $max_datasize;
     }
 
+    $self->_connection_timeout_trigger( $self->connection_timeout );
+    $self->_operation_timeout_trigger( $self->operation_timeout );
+
     my ( $major, $minor ) = $self->_redis->info->{redis_version} =~ /^(\d+)\.(\d+)/;
-    if ( $major < 2 || $major == 2 && $minor <= 4 ) {
+    if ( $major < 2 || $major == 2 && $minor < 8 ) {
         $self->_error( E_REDIS );
-        confess 'Needs Redis server version 2.6 or higher';
+        confess 'Needs Redis server version 2.8 or higher';
     }
+
+    return;
 }
 
 #-- public attributes ----------------------------------------------------------
 
+subtype __PACKAGE__.'::NonNegNum',
+    as 'Num',
+    where { $_ >= 0 },
+    message { format_message( '%s is not a non-negative number!', $_ ) }
+;
+
 =head2 METHODS
 
 The following methods are available for object of the C<Kafka::Producer> class:
@@ -749,6 +815,85 @@ has 'timeout'           => (
     default     => DEFAULT_TIMEOUT,
 );
 
+=head3 reconnect_on_error
+
+Controls ability to force re-connection with Redis on error.
+
+=cut
+has reconnect_on_error      => (
+    is          => 'rw',
+    isa         => 'Bool',
+    default     => 0,
+);
+
+=head3 connection_timeout
+
+Controls socket timeout for Redis server connection, number of seconds (can be fractional).
+
+NOTE: Changes external socket configuration.
+
+=cut
+has connection_timeout      => (
+    is          => 'rw',
+    isa         => 'Maybe['.__PACKAGE__.'::NonNegNum]',
+    default     => undef,
+    trigger     => \&_connection_timeout_trigger,
+);
+
+sub _connection_timeout_trigger {
+    my ( $self, $timeout, $old_timeout ) = @_;
+
+    return if scalar( @_ ) == 2 && ( !defined( $timeout ) && !defined( $old_timeout ) );
+
+    if ( my $redis = $self->_redis ) {
+        my $socket = _INSTANCE( $redis->{sock}, 'IO::Socket' ) or confess( 'Bad socket object' );
+        # IO::Socket provides a way to set a timeout on the socket,
+        # but the timeout will be used only for connection,
+        # not for reading / writing operations.
+        $socket->timeout( $redis->{cnx_timeout} = $timeout );
+    }
+
+    return;
+}
+
+=head3 operation_timeout
+
+Controls socket timeout for Redis server read and write operations, number of seconds (can be fractional).
+
+NOTE: Changes external socket configuration.
+
+=cut
+has operation_timeout       => (
+    is          => 'rw',
+    isa         => 'Maybe['.__PACKAGE__.'::NonNegNum]',
+    default     => undef,
+    trigger     => \&_operation_timeout_trigger,
+);
+
+sub _operation_timeout_trigger {
+    my ( $self, $timeout, $old_timeout ) = @_;
+
+    return if scalar( @_ ) == 2 && ( !defined( $timeout ) && !defined( $old_timeout ) );
+
+    if ( my $redis = $self->_redis ) {
+        my $socket = _INSTANCE( $redis->{sock}, 'IO::Socket' ) or confess( 'Bad socket object' );
+        # IO::Socket::Timeout provides a way to set a timeout
+        # on read / write operations on an IO::Socket instance,
+        # or any IO::Socket::* modules, like IO::Socket::INET.
+        if ( defined $timeout ) {
+            $redis->{write_timeout} = $redis->{read_timeout} = $timeout;
+            $redis->_maybe_enable_timeouts( $socket );
+            $socket->enable_timeout;
+        } else {
+            $redis->{write_timeout} = $redis->{read_timeout} = 0;
+            $redis->_maybe_enable_timeouts( $socket );
+            $socket->disable_timeout;
+        }
+    }
+
+    return;
+}
+
 =head3 C<max_datasize>
 
 Provides access to the C<max_datasize> attribute.
@@ -789,7 +934,7 @@ has 'last_errorcode'    => (
     reader      => 'last_errorcode',
     writer      => '_set_last_errorcode',
     isa         => 'Int',
-    default     => 0,
+    default     => E_NO_ERROR,
 );
 
 #-- private attributes ---------------------------------------------------------
@@ -819,7 +964,7 @@ has '_transaction'      => (
 );
 
 has '_lua_scripts'          => (
-    is          => 'ro',
+    is          => 'rw',
     isa         => 'HashRef[Str]',
     lazy        => 1,
     init_arg    => undef,
@@ -1056,6 +1201,8 @@ sub get_job_data {
 
         return @data;
     }
+
+    return;
 }
 
 =head3 C<get_job_meta_fields( $job )>
@@ -1204,6 +1351,8 @@ sub get_next_job {
     my @keys = ();
     push @keys, map { $self->_qkey( $_ ) } @$queues;
 
+    $self->_error( E_NO_ERROR );
+
     if ( @keys ) {
         @keys = shuffle( @keys );
 
@@ -1276,6 +1425,8 @@ sub _get_next_job {
         # to the next job ID selection
         return;
     }
+
+    return;
 }
 
 =head3 C<update_job( $job )>
@@ -1307,6 +1458,8 @@ sub update_job {
     my @modified = $job->modified_attributes;
     return 0 unless @modified;
 
+    $self->_error( E_NO_ERROR );
+
     my $id = $job->id;
     my $key = $self->_jkey( $id );
     $self->_call_redis( 'WATCH', $key );
@@ -1491,7 +1644,7 @@ If there is no connection to the Redis server (methods return 0), the connection
 In this case, to continue working with the queue,
 you must re-create the C<Redis::JobQueue> object with the L</new> method.
 When using an external connection to the server,
-to check the connection to the server you can use the C<$redis->echo( ... )> call.
+to check the connection to the server you can use the C<$redis-E<gt>echo( ... )> call.
 This is useful to avoid closing the connection to the Redis server unintentionally.
 
 =cut
@@ -1500,14 +1653,11 @@ sub ping {
 
     $self->_error( E_NO_ERROR );
 
-    my $ret;
-    try {
-        $ret = $self->_redis->ping;
-    } catch {
-        $self->_redis_exception( $_ );
-    };
+    my $ret = $self->_redis->ping;
 
-    return( ( $ret // '<undef>' ) eq 'PONG' ? 1 : 0 );
+    $ret = ( $ret // '<undef>' ) eq 'PONG' ? 1 : 0;
+
+    return $ret;
 }
 
 =head3 C<quit>
@@ -1520,8 +1670,8 @@ The following example illustrates use of the C<quit> method:
 
 It does not close the connection to the Redis server if it is an external connection provided
 to queue constructor as existing L<Redis> object.
-When using an external connection (eg, C<$redis = Redis-> new (...);>),
-to close the connection to the Redis server, call C<$redis->quit> after calling this method.
+When using an external connection (eg, C<$redis = Redis-E<gt> new (...);>),
+to close the connection to the Redis server, call C<$redis-E<gt>quit> after calling this method.
 
 =cut
 sub quit {
@@ -1530,14 +1680,7 @@ sub quit {
     return if $] >= 5.14 && ${^GLOBAL_PHASE} eq 'DESTRUCT';
 
     $self->_error( E_NO_ERROR );
-
-    unless ( $self->_use_external_connection ) {
-        try {
-            $self->_redis->quit;
-        } catch {
-            $self->_redis_exception( $_ );
-        };
-    }
+    $self->_redis->quit unless $self->_use_external_connection;
 
     return;
 }
@@ -1606,20 +1749,53 @@ sub queue_status {
         $self->_lua_script_cmd( 'queue_status' ),
         0,
         $maybe_queue,
-        Time::HiRes::time,
+        Time::HiRes::time(),
     );
 
     return \%qstatus;
 }
 
+=head3 C<queue_length>
+
+Gets queue length from the Redis server.
+Queue name is obtained from the argument. The can be a string representing a queue name
+or a L<Redis::JobQueue::Job|Redis::JobQueue::Job> object.
+
+If queue does not exist, it is interpreted as an empty list and 0 is returned.
+
+The following examples illustrate uses of the C<queue_length> method:
+
+    $queue_length = $jq->queue_length( $queue_name );
+    # or
+    $queue_length = $jq->queue_status( $job );
+
+=cut
+sub queue_length {
+    my ( $self, $maybe_queue ) = @_;
+
+    defined( _STRING( $maybe_queue ) ) || _INSTANCE( $maybe_queue, 'Redis::JobQueue::Job' )
+        || confess $self->_error( E_MISMATCH_ARG );
+
+    $maybe_queue = $maybe_queue->queue
+        if ref $maybe_queue;
+
+    $self->_error( E_NO_ERROR );
+
+    my $queue_lenght = $self->_call_redis( 'LLEN', $self->_qkey( $maybe_queue ) );
+
+    return $queue_lenght;
+}
+
 #-- private methods ------------------------------------------------------------
 
 sub _redis_exception {
     my ( $self, $error ) = @_;
 
+    my $err_msg = '';
+
     # Use the error messages from Redis.pm
     if (
-               $error =~ /^Could not connect to Redis server at /
+               $error =~ /Could not connect to Redis server at /    # not in start if not reconnected
             || $error =~ /^Can't close socket: /
             || $error =~ /^Not connected to any server/
             # Maybe for pub/sub only
@@ -1627,6 +1803,12 @@ sub _redis_exception {
             || $error =~ /^Redis server closed connection/
         ) {
         $self->_error( E_NETWORK );
+        $err_msg = $self->_do_reconnect( E_NETWORK, $err_msg ) if !$self->_transaction && $self->reconnect_on_error;
+    } elsif (
+            $error =~ /^\[[^]]+\]\s+NOSCRIPT No matching script. Please use EVAL./
+        ) {
+        $self->_clear_sha1;
+        return 1;
     } elsif (
                $error =~ /[\S+] ERR command not allowed when used memory > 'maxmemory'/
             || $error =~ /[\S+] OOM command not allowed when used memory > 'maxmemory'/
@@ -1634,6 +1816,7 @@ sub _redis_exception {
         $self->_error( E_MAX_MEMORY_LIMIT );
     } else {
         $self->_error( E_REDIS );
+        $err_msg = $self->_do_reconnect( E_REDIS, $err_msg ) if !$self->_transaction && $self->reconnect_on_error;
     }
 
     if ( $self->_transaction ) {
@@ -1642,7 +1825,10 @@ sub _redis_exception {
         };
         $self->_transaction( 0 );
     }
-    die $error;
+
+    die( format_message( '%s %s', $error, $err_msg ) );
+
+    return;
 }
 
 sub _redis_constructor {
@@ -1670,9 +1856,11 @@ sub _call_redis {
     my $self   = shift;
     my $method = shift;
 
-    my @result;
     $self->_error( E_NO_ERROR );
 
+    $self->_wait_used_memory if $self && $method =~ /^EVAL/i;
+
+    my @result;
     my $error;
     # if you use "$method eq 'HSET'" then use $_[0..2] to reduce data copies
     # $_[0] - key
@@ -1739,12 +1927,23 @@ sub _call_redis {
         # For non-serialized fields: UTF8 can not be transferred to the Redis server
         confess format_message( '%s (utf8 in %s)', $self->_error( E_MISMATCH_ARG ), $_[1] );
     } else {
+        my $try_again;
         my @args = @_;
-        try {
-            @result = $self->_redis->$method( @args );
-        } catch {
-            $error = $_;
-        };
+        RUN_METHOD: {
+            try {
+                @result = $self->_redis->$method( @args );
+            } catch {
+                $error = $_;
+                $try_again = $self->_redis_exception( $error );
+            };
+
+            if ( $try_again && $method eq 'EVALSHA' ) {
+                $self->_lua_scripts->{ $_running_script_name } = $args[0];  # sha1
+                $args[0] = $_running_script_body;
+                $method = 'EVAL';
+                redo RUN_METHOD;
+            }
+        }
     }
 
     $self->_redis_exception( $error )
@@ -1767,6 +1966,76 @@ sub _call_redis {
     return wantarray ? @result : $result[0];
 }
 
+sub _wait_used_memory {
+    return unless $WAIT_USED_MEMORY;
+
+    my ( $self ) = @_;
+
+    my ( undef, $maxmemory ) = $self->_call_redis( 'CONFIG', 'GET', 'maxmemory' );
+    my $sleepped;
+    if ( $maxmemory ) {
+        my $max_timeout = $self->operation_timeout || DEFAULT_OPERATION_TIMEOUT;
+        my $timeout = 0.01;
+        $sleepped = 0;
+        WAIT_USED_MEMORY: {
+            my ( $used_memory ) = $self->_call_redis( 'INFO', 'memory' ) =~ /used_memory:(\d+)/;
+            if ( $used_memory < $maxmemory || $sleepped > $max_timeout ) {
+                last WAIT_USED_MEMORY;
+            }
+
+            Time::HiRes::sleep $timeout;
+            $sleepped += $timeout;
+#            $self->_redis->connect;
+            redo WAIT_USED_MEMORY;
+        }
+    }
+
+    return $sleepped;
+}
+
+sub _reconnect {
+    my ( $self ) = @_;
+
+    if ( !$self->_transaction && $self->reconnect_on_error && !$self->ping ) {
+        my $err_msg = $self->_do_reconnect;
+        $self->_redis_exception( $err_msg )
+            if $err_msg;
+    }
+}
+
+sub _do_reconnect {
+    my $self    = shift;
+    my $err     = shift // 0;
+    my $msg     = shift;
+
+    my $err_msg = '';
+    if (
+            !$err || (
+                   $err != E_MISMATCH_ARG
+                && $err != E_DATA_TOO_LARGE
+                && $err != E_MAX_MEMORY_LIMIT
+                && $err != E_JOB_DELETED
+            )
+        ) {
+        try {
+            # NOTE: socket recreated
+            $self->_redis->quit;
+            $self->_redis->connect;
+        } catch {
+            my $error = $_;
+            $err_msg = "(Not reconnected: $error)";
+        };
+    }
+
+    if ( $err_msg ) {
+        $msg = defined( $msg )
+            ? ( $msg ? "$msg " : '' )."($err_msg)"
+            : $err_msg;
+    }
+
+    return $msg;
+}
+
 sub _jkey {
     my $self        = shift;
 
@@ -1793,16 +2062,27 @@ sub _get_job_id {
 sub _lua_script_cmd {
     my ( $self, $name ) = @_;
 
-    my $sha1 = $self->_lua_scripts->{ $name };
+    $_running_script_name = $name;
+    $_running_script_body = $lua_script_body{ $name };
+
+    my $sha1 = $self->_lua_scripts->{ $_running_script_name };
     unless ( $sha1 ) {
-        $sha1 = $self->_lua_scripts->{ $name } = sha1_hex( $lua_script_body{ $name } );
+        $sha1 = $self->_lua_scripts->{ $_running_script_name } = sha1_hex( $_running_script_body );
         unless ( ( $self->_call_redis( 'SCRIPT', 'EXISTS', $sha1 ) )[0] ) {
-            return( 'EVAL', $lua_script_body{ $name } );
+            return( 'EVAL', $_running_script_body );
         }
     }
     return( 'EVALSHA', $sha1 );
 }
 
+sub _clear_sha1 {
+    my ( $self ) = @_;
+
+    $self->_lua_scripts( {} );
+
+    return;
+}
+
 sub _error {
     my ( $self, $error_code ) = @_;
 
@@ -2169,9 +2449,9 @@ store such string as part of L<workload|Redis::JobQueue::Job/workload> / L<resul
 
 =back
 
-Needs Redis server version 2.6 or higher as module uses Redis Lua scripting.
+Needs Redis server version 2.8 or higher as module uses Redis Lua scripting.
 
-The use of C<maxmemory-police all*> in the F<redis.conf> file could lead to
+The use of C<maxmemory-policy all*> in the F<redis.conf> file could lead to
 a serious (but hard to detect) problem as Redis server may delete
 the job queue objects.
 
@@ -2183,6 +2463,8 @@ to forward the request to the appropriate node in the cluster.
 We strongly recommend using of C<maxmemory> option in the F<redis.conf> file if
 the data set may be large.
 
+WARN: Not use C<maxmemory> less than for example 70mb (60 connections) for avoid 'used_memory > maxmemory' problem.
+
 The C<Redis::JobQueue> package was written, tested, and found working on recent
 Linux distributions.
 
@@ -2219,6 +2501,8 @@ L<https://github.com/TrackingSoft/Redis-JobQueue>
 
 Sergey Gladkov, E<lt>sgladkov@trackingsoft.comE<gt>
 
+Please use GitHub project link above to report problems or contact authors.
+
 =head1 CONTRIBUTORS
 
 Alexander Solovey
